import pickle
from datetime import timedelta
import os
import time
import psutil
from typing import Dict, Union, Tuple, Optional

import numpy as np
from tqdm import tqdm

from gluonts.dataset.common import ListDataset
from gluonts.dataset.field_names import FieldName
from gluonts.model.deepar import DeepAREstimator
from gluonts.mx.distribution.neg_binomial import NegativeBinomialOutput
from gluonts.mx.distribution.gaussian import GaussianOutput
from gluonts.mx.trainer import Trainer
from gluonts.evaluation.backtest import make_evaluation_predictions

from htsmodels.results.calculate_metrics import CalculateResultsBottomUp
from htsmodels.utils.logger import Logger
from htsmodels import __version__


class DeepAR:
    def __init__(
        self,
        dataset,
        groups,
        input_dir="./",
        n_samples=200,
        log_dir: str = ".",
    ):
        self.dataset = dataset
        self.groups = groups
        self.timer_start = time.time()
        self.wall_time_train = None
        self.wall_time_predict = None
        self.wall_time_total = None
        self.input_dir = input_dir
        self.stat_cat_cardinalities = [
            v for k, v in self.groups["train"]["groups_n"].items()
        ]
        self.groups['train']['data'] = self.groups['train']['data'].reshape(-1, self.groups['train']['s'])
        self.stat_cat = np.concatenate(
            ([v.reshape(-1, 1) for k, v in self.groups["train"]["groups_idx"].items()]),
            axis=1,
        )
        self.dates = [groups["dates"][0] for _ in range(groups["train"]["s"])]
        self.n_samples = n_samples

        self.time_int = self._determine_time_interval()

        self.n_train = groups["train"]["n"]
        self.n_predict = groups["predict"]["n"]
        self.s = groups["train"]["s"]
        self.h = groups["h"]

        self._initialize_loggers(log_dir)

        self.model_version = __version__

    def _determine_time_interval(self):
        time_interval = (self.groups["dates"][1] - self.groups["dates"][0]).days
        if time_interval < 8:
            return "W"
        elif time_interval < 32:
            return "M"
        elif time_interval < 93:
            return "Q"
        elif time_interval < 367:
            return "Y"

    def _initialize_loggers(self, log_dir):
        self.logger_train = Logger(
            "train",
            algorithm="deepar",
            dataset=self.dataset,
            to_file=True,
            log_dir=log_dir,
        )
        self.logger_predict = Logger(
            "predict",
            algorithm="deepar",
            dataset=self.dataset,
            to_file=True,
            log_dir=log_dir,
        )
        self.logger_metrics = Logger(
            "metrics",
            algorithm="deepar",
            dataset=self.dataset,
            to_file=True,
            log_dir=log_dir,
        )

    def _build_train_ds(self):
        train_target_values = self.groups["train"]["data"].T

        train_ds = ListDataset(
            [
                {
                    FieldName.TARGET: target.reshape(-1),
                    FieldName.START: start,
                    FieldName.FEAT_STATIC_CAT: fsc,
                }
                for (target, start, fsc) in zip(
                    train_target_values, self.dates, self.stat_cat
                )
            ],
            freq=self.time_int,
        )

        return train_ds

    def _build_test_ds(self):
        test_target_values = self.groups["predict"]["data"].reshape(
            self.groups["predict"]["s"], self.groups["predict"]["n"]
        )

        test_ds = ListDataset(
            [
                {
                    FieldName.TARGET: target,
                    FieldName.START: start,
                    FieldName.FEAT_STATIC_CAT: fsc,
                }
                for (target, start, fsc) in zip(
                    test_target_values, self.dates, self.stat_cat
                )
            ],
            freq=self.time_int,
        )

        return test_ds

    def train(self, lr=1e-3, epochs=100, dist="Gaussian"):
        train_ds = self._build_train_ds()

        distribution = self._select_distribution(dist)

        estimator = DeepAREstimator(
            prediction_length=self.groups["h"],
            freq=self.time_int,
            distr_output=distribution,
            use_feat_dynamic_real=False,
            use_feat_static_cat=True,
            cardinality=self.stat_cat_cardinalities,
            trainer=Trainer(learning_rate=lr, epochs=epochs, num_batches_per_epoch=50),
        )

        model = estimator.train(train_ds)

        self.wall_time_train = time.time() - self.timer_start
        td = timedelta(seconds=int(time.time() - self.timer_start))
        self.logger_train.info(f"wall time train {str(td)}")
        return model

    @staticmethod
    def _select_distribution(dist):
        if dist == "NegativeBinomial":
            return NegativeBinomialOutput()
        else:
            return GaussianOutput()

    def predict(self, model, track_mem=True):
        timer_start = time.time()
        test_ds = self._build_test_ds()

        forecast_it, ts_it = make_evaluation_predictions(
            dataset=test_ds, predictor=model, num_samples=self.n_samples
        )

        if track_mem:
            # Track RAM usage
            process = psutil.Process(os.getpid())
            mem = process.memory_info().rss / (1024**3)
            self.logger_predict.info(f"predict used {mem:.3f} GB of RAM")

        print("Obtaining time series predictions ...")
        forecasts = list(tqdm(forecast_it, total=len(test_ds)))

        pred_mean = np.zeros((self.n_predict, self.s))
        pred_std = np.zeros((self.n_predict, self.s))

        for ts in range(len(test_ds)):
            pred_mean[:, ts] = np.concatenate(
                (self.groups["train"]["data"][:, ts], forecasts[ts].mean), axis=0
            )
            pred_std[:, ts] = np.concatenate(
                (np.zeros((self.n_train,)), np.std(forecasts[ts].samples, axis=0)),
                axis=0,
            )

        self.wall_time_predict = time.time() - timer_start
        return pred_mean, pred_std

    @staticmethod
    def _validate_param(param, valid_values):
        if param not in valid_values:
            raise ValueError(f"{param} is not a valid value")

    def store_results(
        self,
        res: np.ndarray,
        res_type: str,
        res_measure: str,
        track_mem: bool = True,
    ):
        """
        Store results

        Parameters:
            res: np array with the results with shape (n,s) - note that n depends of the res_type
            res_type: defines the type of results, could be 'fit_pred' to receive fitted values plus
                predictions or 'pred' to only store predictions
            res_measure: defines the measure to store, could be 'mean' or 'std'

        Returns:
            numpy.ndarray: Array of shape (n_samples, n_prediction_points, n_groups)
                containing the prediction samples.
        """
        """
        Store results, res_type should be used to define the type of results,
        could be 'fit_pred' to receive fitted values plus predictions or 'pred'
        to only store predictions
        """
        self._validate_param(res_type, ["fitpred", "pred"])
        self._validate_param(res_measure, ["mean", "std"])
        with open(
            f"{self.input_dir}results_{res_type}_{res_measure}_gp_cov_{self.dataset}_{self.model_version}.pickle",
            "wb",
        ) as handle:
            if track_mem:
                process = psutil.Process(os.getpid())
                mem = process.memory_info().rss / (1024**3)
                self.logger_metrics.info(f"Storing results used {mem:.3f} GB of RAM")
            pickle.dump(res, handle, pickle.HIGHEST_PROTOCOL)

    def store_metrics(
        self,
        res: Dict[str, Dict[str, Union[float, np.ndarray]]],
        track_mem: bool = True,
    ):
        with open(
            f"{self.input_dir}metrics_gp_cov_{self.dataset}_{self.model_version}.pickle",
            "wb",
        ) as handle:
            if track_mem:
                process = psutil.Process(os.getpid())
                mem = process.memory_info().rss / (1024**3)
                self.logger_metrics.info(
                    f"Storing error metrics used {mem:.3f} GB of RAM"
                )
            pickle.dump(res, handle, pickle.HIGHEST_PROTOCOL)

    def metrics(
        self,
        pred_mean: np.ndarray,
        pred_std: np.ndarray,
        track_mem: bool = True,
    ):
        calc_results = CalculateResultsBottomUp(
            predictions_mean=pred_mean,
            predictions_std=pred_std,
            groups=self.groups,
            dataset=self.dataset,
        )
        res = calc_results.calculate_metrics()
        if track_mem:
            process = psutil.Process(os.getpid())
            mem = process.memory_info().rss / (1024**3)
            self.logger_metrics.info(
                f"calculating error metrics used {mem:.3f} GB of RAM"
            )
        self.wall_time_total = time.time() - self.timer_start

        res["wall_time"] = {}
        res["wall_time"]["wall_time_train"] = self.wall_time_train
        res["wall_time"]["wall_time_predict"] = self.wall_time_predict
        res["wall_time"]["wall_time_total"] = self.wall_time_total

        return res



import unittest
import tsaugmentation as tsag
from htsmodels.models.mint import MinT
from htsmodels.preprocessing.subsampled_dataset import update_missing_values


class TestModel(unittest.TestCase):

    def setUp(self):
        self.data = tsag.preprocessing.CreateGroups('prison', 0.75).read_subsampled_groups()
        self.data_w_missing_values = update_missing_values(self.data)
        self.mint = MinT(dataset='prison', groups=self.data_w_missing_values)

    def test_correct_train(self):
        model = self.mint.train()
        self.assertIsNotNone(model)




import numpy as np
from sklearn.metrics import mean_squared_error
import properscoring as ps
from sktime.performance_metrics.forecasting import MeanAbsoluteScaledError


class CalculateResultsBase:
    """Calculate the results and store them using pickle files

    Currently we have implemented MASE and RMSE.

    Attributes:
        pred_samples (array): predictions of shape [number of samples, h, number of series]
            - we transform it immediately to the shape [h, number of series] by averaging over the samples
        groups (obj): dict containing the data and several other attributes of the dataset

    """

    def __init__(self, groups, dataset):
        self.groups = groups
        self.seas = self.groups["seasonality"]
        self.h = self.groups["h"]
        self.n = self.groups["predict"]["n"]
        self.s = self.groups["predict"]["s"]
        self.y_f = self.groups["predict"]["data"].reshape(self.s, self.n).T
        self.errs = ["mase", "rmse", "CRPS"]
        self.levels = list(self.groups["train"]["groups_names"].keys())
        self.levels.extend(("bottom", "total"))
        self.dataset = dataset
        self.mase = MeanAbsoluteScaledError(multioutput="raw_values")

    def calculate_metrics_for_individual_group(
        self, group_name, y, predictions_mean, predictions_std, error_metrics
    ):
        """Calculates the main metrics for each group

        Args:
            param group_name: group that we want to calculate the error metrics
            param y: original series values with the granularity of the group to calculate
            param predictions_mean: predictions mean with the granularity of the group to calculate
            param error_metrics: dict to add new results
            param predictions_sample: samples of the predictions
            param predictions_variance: variance of the predictions

        Returns:
            error (obj): contains both the error metric for each individual series of each group and the average

        """

        y_true = y[-self.h :, :]
        y_train = y[: -self.h, :]
        f = predictions_mean[-self.h :]
        f_std = predictions_std[-self.h :]
        error_metrics["mase"][f"{group_name}_ind"] = np.round(
            self.mase(y_true=y_true, y_pred=f, y_train=y_train, sp=self.seas), 3
        )
        error_metrics["mase"][f"{group_name}"] = np.round(
            np.mean(error_metrics["mase"][f"{group_name}_ind"]), 3
        )
        error_metrics["rmse"][f"{group_name}_ind"] = np.round(
            mean_squared_error(y_true, f, squared=False, multioutput="raw_values"),
            3,
        )
        error_metrics["rmse"][f"{group_name}"] = np.round(
            np.mean(error_metrics["rmse"][f"{group_name}_ind"]), 3
        )

        error_metrics["CRPS"][f"{group_name}"] = ps.crps_gaussian(
            y_true, f, f_std
        ).mean()
        error_metrics["CRPS"][f"{group_name}_ind"] = ps.crps_gaussian(
            y_true, f, f_std
        ).mean(axis=0)

        return error_metrics


class CalculateResultsBottomUp(CalculateResultsBase):
    r"""
    Calculate results for the bottom-up strategy.

    From the prediction of the bottom level series, aggregate the results for the upper levels
    considering the hierarchical structure and compute the error metrics accordingly.

    Parameters
    ----------
    pred_samples : numpy array
        results for the bottom series
    groups : dict
        all the information regarding the different groups
    """

    def __init__(self, groups, dataset, predictions_mean, predictions_std):
        super().__init__(groups=groups, dataset=dataset)
        self.groups_names = list(self.groups["predict"]["groups_names"].keys())
        self.n_samples = groups["predict"]["n"]
        self.predictions_mean = predictions_mean
        self.predictions_std = predictions_std

    def compute_error_for_every_group(self, error_metrics):
        """Computes the error metrics for all the groups

        Returns:
            error (obj): - contains all the error metric for each group in the dataset
                         - contains all the predictions for all the groups

        """
        group_element_active = dict()
        for group in list(self.groups["predict"]["groups_names"].keys()):
            n_elements_group = self.groups["predict"]["groups_names"][group].shape[0]
            group_elements = self.groups["predict"]["groups_names"][group]
            groups_idx = self.groups["predict"]["groups_idx"][group]

            y_g = np.zeros((self.n_samples, n_elements_group))
            mean_g = np.zeros((self.n_samples, n_elements_group))
            std_g = np.zeros((self.n_samples, n_elements_group))

            for group_idx, element_name in enumerate(group_elements):
                group_element_active[element_name] = np.where(
                    groups_idx == group_idx, 1, 0
                ).reshape((1, -1))

                y_g[:, group_idx] = np.sum(
                    group_element_active[element_name] * self.y_f, axis=1
                )
                mean_g[:, group_idx] = np.sum(
                    group_element_active[element_name] * self.predictions_mean, axis=1
                )
                std_g[:, group_idx] = np.sum(
                    group_element_active[element_name] * self.predictions_mean, axis=1
                )

            error_metrics = self.calculate_metrics_for_individual_group(
                group_name=group,
                y=y_g,
                predictions_mean=mean_g,
                predictions_std=std_g,
                error_metrics=error_metrics,
            )
        return error_metrics

    def bottom_up(self, level, error_metrics):
        """Aggregates the results for all the groups

        Returns:
            error (obj): - contains all the error metric for the specific level

        """
        if level == "bottom":
            error_metrics = self.calculate_metrics_for_individual_group(
                group_name=level,
                y=self.y_f,
                predictions_mean=self.predictions_mean,
                predictions_std=self.predictions_std,
                error_metrics=error_metrics,
            )
        elif level == "total":
            np.sqrt(np.sum(self.predictions_std**2, axis=1)).reshape(-1, 1)
            error_metrics = self.calculate_metrics_for_individual_group(
                group_name=level,
                y=np.sum(self.y_f, axis=1).reshape(-1, 1),
                predictions_mean=np.sum(self.predictions_mean, axis=1).reshape(-1, 1),
                # The variance of the resulting distribution will be the sum
                # of the variances of the original Gaussian distributions
                predictions_std=np.sqrt(
                    np.sum(self.predictions_std**2, axis=1)
                ).reshape(-1, 1),
                error_metrics=error_metrics,
            )
        elif level == "groups":
            self.compute_error_for_every_group(error_metrics)

        return error_metrics

    def calculate_metrics(self):
        """Aggregates the results for all the groups

        Returns:
            error (obj): - contains all the error metric for each individual series of each group and the average
                         - contains all the predictions for all the series and groups

        """
        error_metrics = dict()
        error_metrics["mase"] = {}
        error_metrics["rmse"] = {}
        error_metrics["CRPS"] = {}

        error_metrics = self.bottom_up("bottom", error_metrics)
        error_metrics = self.bottom_up("total", error_metrics)
        error_metrics = self.bottom_up("groups", error_metrics)

        # Aggregate all errors and create the 'all' category
        for err in self.errs:
            error_metrics[err]["all_ind"] = np.squeeze(
                np.concatenate(
                    [
                        error_metrics[err][f"{x}_ind"].reshape((-1, 1))
                        for x in self.levels
                    ],
                    0,
                )
            )
            error_metrics[err]["all"] = np.mean(error_metrics[err]["all_ind"])

        return error_metrics


class CalculateResultsMint(CalculateResultsBase):
    r"""
    Calculate results for MinT reconciliation.

    The usage is nearly identical to the bottom-up reconciliation strategy, but the implementation
    is a bit different, since the MinT algorithm yields results for all levels in the hierarchy.
    We just need to filter those to get the mean, lower and upper values for each level.

    Parameters
    ----------
    df_results_mint : pandas
        df with the results
    groups : dict
        dictionary with all the information regarding the different groups
    """

    def __init__(self, df_results_mint, groups, dataset):
        super().__init__(groups=groups, dataset=dataset)
        self.df_results_mint = df_results_mint

    def compute_error_for_every_group(self, error_metrics):
        """Computes the error metrics for all the groups

        Returns:
            error (obj): - contains all the error metric for each group in the dataset
                         - contains all the predictions for all the groups

        """
        idx_dict_new = dict()
        for group in list(self.groups["train"]["groups_names"].keys()):
            y_g = np.zeros(
                (
                    self.groups["predict"]["n"],
                    self.groups["train"]["groups_names"][group].shape[0],
                )
            )
            f_g = np.zeros(
                (self.h, self.groups["train"]["groups_names"][group].shape[0])
            )
            std_g = np.zeros(
                (self.h, self.groups["train"]["groups_names"][group].shape[0])
            )

            for idx, name in enumerate(self.groups["train"]["groups_names"][group]):
                idx_dict_new[name] = np.where(
                    self.groups["train"]["groups_idx"][group] == idx, 1, 0
                )

                y_g[:, idx] = np.sum(idx_dict_new[name] * self.y_f, axis=1)
                pred_samples_group = self.df_results_mint.loc[
                    self.df_results_mint[group].isin([name])
                ].drop([group], axis=1)
                f_g[:, idx] = np.asarray(
                    pred_samples_group.loc[
                        pred_samples_group.iloc[
                            :, : self.groups["train"]["g_number"] - 1
                        ]
                        .isin(["<aggregated>"])
                        .all(axis=1)
                    ]["mean"]
                ).reshape((self.h,))
                std_g[:, idx] = np.asarray(
                    pred_samples_group.loc[
                        pred_samples_group.iloc[
                            :, : self.groups["train"]["g_number"] - 1
                        ]
                        .isin(["<aggregated>"])
                        .all(axis=1)
                    ]["std"]
                ).reshape((self.h,))

            error_metrics = self.calculate_metrics_for_individual_group(
                group, y_g, f_g, std_g, error_metrics
            )

        return error_metrics

    def mint_reconciliation(self, level, error_metrics):
        """Get the results for each of the level of aggregation using the MinT strategy (bottom-up not needed)

        Returns:
            error (obj): - contains all the error metric for the specific level

        """
        if level == "bottom":
            n_s = self.s
            pred_samples_mean = (
                np.asarray(
                    self.df_results_mint.loc[
                        ~self.df_results_mint.isin(["<aggregated>"]).any(axis=1)
                    ]["mean"]
                )
                .reshape((n_s, self.h))
                .T
            )
            pred_samples_std = (
                np.asarray(
                    self.df_results_mint.loc[
                        ~self.df_results_mint.isin(["<aggregated>"]).any(axis=1)
                    ]["std"]
                )
                .reshape((n_s, self.h))
                .T
            )
            error_metrics = self.calculate_metrics_for_individual_group(
                level,
                self.y_f,
                pred_samples_mean,
                pred_samples_std,
                error_metrics,
            )

        elif level == "total":
            n_s = 1
            pred_samples_mean = (
                np.asarray(
                    self.df_results_mint.loc[
                        self.df_results_mint.iloc[:, : self.groups["train"]["g_number"]]
                        .isin(["<aggregated>"])
                        .all(axis=1)
                    ]["mean"]
                )
                .reshape((n_s, self.h))
                .T
            )
            pred_samples_std = (
                np.asarray(
                    self.df_results_mint.loc[
                        self.df_results_mint.iloc[:, : self.groups["train"]["g_number"]]
                        .isin(["<aggregated>"])
                        .all(axis=1)
                    ]["std"]
                )
                .reshape((n_s, self.h))
                .T
            )

            error_metrics = self.calculate_metrics_for_individual_group(
                level,
                np.sum(self.y_f, axis=1).reshape(-1, 1),
                pred_samples_mean,
                pred_samples_std,
                error_metrics,
            )
        elif level == "groups":
            self.compute_error_for_every_group(error_metrics)

        return error_metrics

    def calculate_metrics(self):
        """Aggregates the results for all the groups

        Returns:
            error (obj): - contains all the error metric for each individual series of each group and the average
                         - contains all the predictions for all the series and groups

        """
        error_metrics = dict()
        error_metrics["mase"] = {}
        error_metrics["rmse"] = {}
        error_metrics["CRPS"] = {}

        error_metrics = self.mint_reconciliation("bottom", error_metrics)
        error_metrics = self.mint_reconciliation("total", error_metrics)
        error_metrics = self.mint_reconciliation("groups", error_metrics)

        # Aggregate all errors and create the 'all' category
        for err in self.errs:
            error_metrics[err]["all_ind"] = np.squeeze(
                np.concatenate(
                    [
                        error_metrics[err][f"{x}_ind"].reshape((-1, 1))
                        for x in self.levels
                    ],
                    0,
                )
            )
            error_metrics[err]["all"] = np.mean(error_metrics[err]["all_ind"])

        return error_metrics



import logging
import os
import time
from htsmodels import __version__


class Logger:
    def __init__(
        self,
        name,
        dataset,
        algorithm,
        to_file=None,
        log_level=logging.INFO,
        log_dir=".",
    ):
        timestamp = time.strftime("%Y%m%d-%H%M%S", time.gmtime())

        self.logger = logging.getLogger(name)
        self.logger.setLevel(log_level)

        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )

        handler = logging.StreamHandler()  # Outputs the log to the console
        if to_file:
            log_file = os.path.join(
                log_dir,
                "logs",
                f"{algorithm}_{__version__}_{dataset}_log_{timestamp}.txt",
            )
            handler = logging.FileHandler(log_file)  # Outputs the log to a file
        handler.setFormatter(formatter)

        self.logger.addHandler(handler)

    def info(self, message):
        self.logger.info(message)

    def warning(self, message):
        self.logger.warning(message)

    def error(self, message):
        self.logger.error(message)



import unittest
import tsaugmentation as tsag
from htsmodels.models.mint import MinT


class TestModel(unittest.TestCase):

    def setUp(self):
        self.data = tsag.preprocessing.CreateGroups('m5').read_original_groups()
        self.n = self.data['predict']['n']
        self.s = self.data['train']['s']
        self.mint = MinT(dataset='m5', groups=self.data)

    def test_correct_train(self):
        model = self.mint.train()
        self.assertIsNotNone(model)





import pickle
from datetime import datetime, timedelta
import os
import time
import psutil
from typing import Dict, Union, Tuple, Optional

import pandas as pd
import numpy as np
import rpy2.robjects as robjects
from rpy2.robjects import pandas2ri
import rpy2.robjects as ro
from rpy2.robjects.conversion import localconverter

from htsmodels.results.calculate_metrics import CalculateResultsBottomUp
from htsmodels.utils.logger import Logger
from htsmodels import __version__


class MinT:
    def __init__(
        self, dataset, groups, aggregate_key=None, input_dir="./", log_dir: str = "."
    ):

        self.dataset = dataset
        self.groups = groups
        self.timer_start = time.time()
        self.wall_time_train = None
        self.wall_time_predict = None
        self.wall_time_total = None
        self.input_dir = input_dir
        dict_groups = {
            k: np.tile(
                groups["train"]["groups_names"][k][groups["train"]["groups_idx"][k]],
                (groups["predict"]["n"], 1),
            ).T.reshape(
                -1,
            )
            for k in [k for k, v in groups["train"]["groups_n"].items()]
        }
        train_data = groups["train"]["data"].reshape(groups["train"]["n"], -1)
        groups["predict"]["data_matrix"][: groups["train"]["n"], :] = train_data

        dict_groups["Count"] = groups["predict"]["data_matrix"].T.reshape(
            -1,
        )
        dict_groups["Date"] = np.tile(
            np.array(groups["dates"]), (groups["train"]["s"],)
        )
        self.df = pd.DataFrame(dict_groups)
        self.algorithm = "mint"
        self.groups_idx = self.groups["train"]["groups_idx"]
        self.groups_ele_names = self.groups["train"]["groups_names"]
        self.n_predict = self.groups["predict"]["n"]
        self.n_train = self.groups["train"]["n"]
        self.s = self.groups["predict"]["s"]
        self.h = self.groups["h"]
        self.groups_names = list(self.groups["train"]["groups_n"].keys())

        time_interval = (self.groups["dates"][1] - self.groups["dates"][0]).days
        if time_interval < 2:
            self.time_int = "day"
        elif time_interval < 8:
            self.time_int = "week"
        elif time_interval < 32:
            self.time_int = "month"
        elif time_interval < 93:
            self.time_int = "quarter"
        elif time_interval < 367:
            self.time_int = "year"

        self.last_train_date = (
            self.groups["dates"][self.groups["train"]["n"] - 1]
        ).strftime("%Y-%m-%d")
        if aggregate_key:
            self.aggregate_key = aggregate_key
        else:
            # Default is to assume that there is a group structure (e.g. State * Gender)
            # and no direct hierarchy (e.g. State / Region)
            self.aggregate_key = " * ".join(
                [k for k in self.groups["train"]["groups_names"]]
            )

        self.logger_train = Logger(
            "train",
            algorithm="mint",
            dataset=self.dataset,
            to_file=True,
            log_dir=log_dir,
        )
        self.logger_predict = Logger(
            "predict",
            algorithm="mint",
            dataset=self.dataset,
            to_file=True,
            log_dir=log_dir,
        )
        self.logger_metrics = Logger(
            "metrics",
            algorithm="mint",
            dataset=self.dataset,
            to_file=True,
            log_dir=log_dir,
        )

        self.model_version = __version__

    def train(self, algorithm="ets", rec_method="mint", track_mem=True):
        """Train ETS or ARIMA with conventional bottom-up or MinT reconciliation strategies

        :param algorithm: ets or arima
        :param rec_method: reconciliation method (bottom_up, mint or base)
            base means that we just forecast all time series without any reconciliation
        :return: df with results
        """
        robjects.r(
            """
            library('fpp3')
            results_fn <- function(df, time, h, string_aggregate, start_predict_date, algorithm, rec_method) {
              if (time == 'quarter') {
                fn = yearquarter
              } else if (time == 'month') {
                fn = yearmonth
              } else if (time == 'week') {
                fn = yearweek
              } else if (time == 'year') {
                fn = year
              } else if (time == 'day') {
                fn = ymd
              } 
              
              # Get the correct algorithm and assign the respective function
              if (algorithm == 'ets') {
                algo_fn = ETS
              } else if (algorithm == 'arima') {
                algo_fn = ARIMA
              }
              
              # Get the correct reconciliation method and assign the respective function
              if (rec_method == 'bottom_up') {
                rec_method_fn = bottom_up
              } else if (rec_method == 'mint') {
                rec_method_fn = min_trace
              }
              
              data <- df %>%
                mutate(Time = fn(Date)) %>%
                select(-Date) %>%
                as_tsibble(key = colnames(df)[2:length(colnames(df))-2], index = Time) %>%
                relocate(Time)
              
              data_gts <- data %>%
                aggregate_key(.spec = !!rlang::parse_expr(string_aggregate), Count = sum(Count, na.rm = TRUE))              
              fit <- data_gts %>%
                filter(Time <= fn(as.Date(start_predict_date))) %>%
                  model(base = algo_fn(Count)) %>%
                  reconcile(
                    if (rec_method == 'mint') {
                        rec = rec_method_fn(base, method = 'mint_shrink')
                    } else if (rec_method == 'bottom_up') {
                        rec = rec_method_fn(base)
                    } else {
                    }
                  )
              fc <- fit %>% forecast(h = h)
              
              fc_csv = fc %>% 
                        # getting the 95% interval
                        hilo(level = 95) %>%
                        unpack_hilo('95%') %>% 
                        as_tibble %>% 
                        mutate(mean=mean(Count)) %>%
                        mutate(std=sqrt(distributional::variance(Count))) %>%
                        select(-Count) %>% 
                        rename(lower='95%_lower') %>%
                        rename(upper='95%_upper') %>%
                        select(-.mean) %>%
                        rename(time=Time) %>%
                        lapply(as.character) %>% 
                        data.frame(stringsAsFactors=FALSE)
              
              return (fc_csv)
            }
        """
        )
        function_r = robjects.globalenv["results_fn"]
        with localconverter(ro.default_converter + pandas2ri.converter):
            df_r = ro.conversion.py2rpy(self.df)

        df_result_r = function_r(
            df_r,
            self.time_int,
            self.groups["h"],
            self.aggregate_key,
            self.last_train_date,
            algorithm,
            rec_method,
        )
        with localconverter(ro.default_converter + pandas2ri.converter):
            df_result = ro.conversion.rpy2py(df_result_r)

        # When it runs the reconciliation it also runs the base algorithm, so
        # it has two different outputs stored in the df that we need to filter
        if not rec_method == "base":
            df_result = df_result.loc[df_result[".model"] != "base"]

        # name properly the models in the dataframe
        df_result.loc[:, ".model"] = algorithm + "_" + rec_method

        df_result[["mean"]] = df_result[["mean"]].astype("float")
        df_result[["lower"]] = df_result[["lower"]].astype("float")
        df_result[["upper"]] = df_result[["upper"]].astype("float")
        df_result[["std"]] = df_result[["std"]].astype("float")

        if track_mem:
            # Track RAM usage
            process = psutil.Process(os.getpid())
            mem = process.memory_info().rss / (1024**3)
            self.logger_train.info(f"train used {mem:.3f} GB of RAM")

        self.wall_time_train = time.time() - self.timer_start
        td = timedelta(seconds=int(time.time() - self.timer_start))
        self.logger_train.info(f"wall time train {str(td)}")
        self.wall_time_predict = 0

        return df_result

    def predict(self, df_train_results: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        ordered_mean = np.zeros((self.h, self.s))
        ordered_std = np.zeros((self.h, self.s))

        for s in range(self.s):
            df_temp = df_train_results.copy()
            for group in self.groups_names:
                df_temp = df_temp.loc[
                    df_temp[group]
                    == self.groups_ele_names[group][self.groups_idx[group]][s]
                ]

            ordered_mean[:, s] = df_temp["mean"].to_numpy()
            ordered_std[:, s] = df_temp["std"].to_numpy()

        pred_mean = np.concatenate((self.groups["train"]["data"], ordered_mean), axis=0)
        pred_std = np.concatenate(
            (np.zeros((self.n_train, self.s)), ordered_std),
            axis=0,
        )

        return pred_mean, pred_std

    def results(self, pred_mint):
        cols = list(self.groups["train"]["groups_names"].keys())

        if self.time_int == "day":
            pred_mint["Date"] = pred_mint["time"]
        elif self.time_int == "week":
            pred_mint["Date"] = (
                pred_mint["time"]
                .apply(
                    lambda x: datetime.strptime(x + str(0), "%Y W%W%w").strftime(
                        "%Y-%m-%d"
                    )
                )
                .apply(pd.to_datetime)
            )
        elif self.time_int == "month":
            pred_mint["Date"] = (
                pred_mint["time"]
                .apply(lambda x: datetime.strptime(x, "%Y %b").strftime("%Y-%m-%d"))
                .apply(pd.to_datetime)
            )
        elif self.time_int == "quarter":
            pred_mint["Date"] = pd.PeriodIndex(
                pred_mint["time"].str.replace(" ", "-"), freq="Q"
            ).to_timestamp()
        elif self.time_int == "year":
            pred_mint["Date"] = pd.PeriodIndex(
                pred_mint["time"], freq="Y"
            ).to_timestamp()

        cols.append("Date")
        # Zip can sometimes have the dtype int and breaks
        pred_mint = pred_mint.astype({k: "string" for k in cols})
        pred_mint["Date"] = pd.to_datetime(pred_mint["Date"])

        # Filter only the predictions
        pred_mint = pred_mint[pred_mint["Date"] > self.last_train_date]

        return pred_mint

    @staticmethod
    def _validate_param(param, valid_values):
        if param not in valid_values:
            raise ValueError(f"{param} is not a valid value")

    def store_results(
        self,
        res: np.ndarray,
        res_type: str,
        res_measure: str,
        track_mem: bool = True,
    ):
        """
        Store results

        Parameters:
            res: np array with the results with shape (n,s) - note that n depends of the res_type
            res_type: defines the type of results, could be 'fit_pred' to receive fitted values plus
                predictions or 'pred' to only store predictions
            res_measure: defines the measure to store, could be 'mean' or 'std'

        Returns:
            numpy.ndarray: Array of shape (n_samples, n_prediction_points, n_groups)
                containing the prediction samples.
        """
        """
        Store results, res_type should be used to define the type of results,
        could be 'fit_pred' to receive fitted values plus predictions or 'pred'
        to only store predictions
        """
        self._validate_param(res_type, ["fitpred", "pred"])
        self._validate_param(res_measure, ["mean", "std"])
        with open(
            f"{self.input_dir}results_{res_type}_{res_measure}_gp_cov_{self.dataset}_{self.model_version}.pickle",
            "wb",
        ) as handle:
            if track_mem:
                process = psutil.Process(os.getpid())
                mem = process.memory_info().rss / (1024**3)
                self.logger_metrics.info(f"Storing results used {mem:.3f} GB of RAM")
            pickle.dump(res, handle, pickle.HIGHEST_PROTOCOL)

    def store_metrics(
        self,
        res: Dict[str, Dict[str, Union[float, np.ndarray]]],
        track_mem: bool = True,
    ):
        with open(
            f"{self.input_dir}metrics_gp_cov_{self.dataset}_{self.model_version}.pickle",
            "wb",
        ) as handle:
            if track_mem:
                process = psutil.Process(os.getpid())
                mem = process.memory_info().rss / (1024**3)
                self.logger_metrics.info(
                    f"Storing error metrics used {mem:.3f} GB of RAM"
                )
            pickle.dump(res, handle, pickle.HIGHEST_PROTOCOL)

    def metrics(
        self,
        pred_mean: np.ndarray,
        pred_std: np.ndarray,
        track_mem: bool = True,
    ):
        calc_results = CalculateResultsBottomUp(
            predictions_mean=pred_mean,
            predictions_std=pred_std,
            groups=self.groups,
            dataset=self.dataset,
        )
        res = calc_results.calculate_metrics()
        if track_mem:
            process = psutil.Process(os.getpid())
            mem = process.memory_info().rss / (1024**3)
            self.logger_metrics.info(
                f"calculating error metrics used {mem:.3f} GB of RAM"
            )
        self.wall_time_total = time.time() - self.timer_start

        res["wall_time"] = {}
        res["wall_time"]["wall_time_train"] = self.wall_time_train
        res["wall_time"]["wall_time_predict"] = self.wall_time_predict
        res["wall_time"]["wall_time_total"] = self.wall_time_total

        return res




import unittest
import pickle
import tsaugmentation as tsag
from htsmodels.models.mint import MinT
from htsmodels import __version__


class TestModel(unittest.TestCase):

    def setUp(self):
        self.dataset_name = "prison"
        self.data = tsag.preprocessing.PreprocessDatasets(self.dataset_name).apply_preprocess()
        self.n = self.data['predict']['n']
        self.s = self.data['train']['s']
        self.h = self.data['h']
        self.version = __version__
        self.mint = MinT(dataset=self.dataset_name, groups=self.data, input_dir='./results/mint/')
        self.mint.input_dir = f"./results/mint/"

    def test_correct_train(self):
        model = self.mint.train()
        self.assertIsNotNone(model)

    def test_results_interval(self):
        forecasts = self.mint.train()
        results = self.mint.results(forecasts)
        pred_mean, pred_std = self.mint.predict(results)
        res = self.mint.metrics(pred_mean, pred_std)
        self.assertLess(res['mase']['bottom'], 2.2)
        self.assertLess(res['CRPS']['bottom_ind'][0], 5)

    def test_store_pred_std_results_gp(self):
        forecasts = self.mint.train()
        results = self.mint.results(forecasts)
        pred_mean, pred_std = self.mint.predict(results)
        res_type = "pred"
        res_measure = "std"
        self.mint.store_results(
            pred_std[self.n - self.h :, :], res_type=res_type, res_measure=res_measure
        )
        with open(
            f"{self.mint.input_dir}results_{res_type}_{res_measure}_gp_cov_{self.dataset_name}_{self.version}.pickle",
            "rb",
        ) as handle:
            res = pickle.load(handle)
        self.assertTrue(res.shape == (self.h, self.s))

    def test_store_metrics_mint(self):
        forecasts = self.mint.train()
        results = self.mint.results(forecasts)
        pred_mean, pred_std = self.mint.predict(results)
        res = self.mint.metrics(pred_mean, pred_std)
        self.mint.store_metrics(res)
        with open(
            f"{self.mint.input_dir}metrics_gp_cov_{self.dataset_name}_{self.version}.pickle",
            "rb",
        ) as handle:
            res = pickle.load(handle)
        keys = list(res.keys())
        self.assertTrue(keys == ['mase', 'rmse', 'CRPS', 'wall_time'])



import unittest
import tsaugmentation as tsag
from htsmodels.models.mint import MinT


class TestModel(unittest.TestCase):

    def setUp(self):
        self.data = tsag.preprocessing.PreprocessDatasets('tourism', test_size=228*10).apply_preprocess()
        self.n = self.data['predict']['n']
        self.s = self.data['train']['s']
        self.mint = MinT(dataset='tourism', groups=self.data)

    def test_correct_train(self):
        model = self.mint.train()
        self.assertIsNotNone(model)

    def test_results_interval(self):
        forecasts = self.mint.train()
        results = self.mint.results(forecasts)
        pred_mean, pred_std = self.mint.predict(results)
        res = self.mint.metrics(pred_mean, pred_std)
        self.assertLess(res['mase']['bottom'], 2.2)
        self.assertLess(res['CRPS']['bottom_ind'][0], 100)



import unittest
import pickle
import tsaugmentation as tsag
from htsmodels.models.standard_gp import SGP
from htsmodels import __version__


class TestModel(unittest.TestCase):

    def setUp(self):
        self.dataset_name = 'prison'
        self.data = tsag.preprocessing.PreprocessDatasets(self.dataset_name).apply_preprocess()
        self.n = self.data['predict']['n']
        self.s = self.data['train']['s']
        self.h = self.data['h']
        self.version = __version__
        self.gpf = SGP('prison', self.data)

    def test_results_interval_piecewise(self):
        model, like = self.gpf.train(n_iterations=100)
        pred, pred_scaled = self.gpf.predict(model, like)
        res = self.gpf.metrics(pred[0], pred[1])
        self.assertLess(res['mase']['bottom'], 5)

    def test_results_interval_linear(self):
        model, like = self.gpf.train(n_iterations=100, mean_function='linear')
        pred, pred_scaled = self.gpf.predict(model, like)
        res = self.gpf.metrics(pred[0], pred[1])
        self.assertLess(res['mase']['bottom'], 5)

    def test_results_interval_zero(self):
        model, like = self.gpf.train(n_iterations=100, mean_function='zero')
        pred, pred_scaled = self.gpf.predict(model, like)
        res = self.gpf.metrics(pred[0], pred[1])
        self.assertLess(res['mase']['bottom'], 5)

    def test_store_pred_std_results_gp(self):
        model, like = self.gpf.train(n_iterations=100, mean_function='zero')
        preds, preds_scaled = self.gpf.predict(model, like)
        res_type = "pred"
        res_measure = "std"
        self.gpf.store_results(
            preds[1][self.n - self.h :, :], res_type=res_type, res_measure=res_measure
        )
        with open(
            f"{self.gpf.input_dir}results_{res_type}_{res_measure}_gp_cov_{self.dataset_name}_{self.version}.pickle",
            "rb",
        ) as handle:
            res = pickle.load(handle)
        self.assertTrue(res.shape == (self.h, self.s))

    def test_store_metrics_gp(self):
        model, like = self.gpf.train(n_iterations=100, mean_function='zero')
        pred, pred_scaled = self.gpf.predict(model, like)
        self.gpf.input_dir = f"./results/gpf/"
        preds, preds_scaled = self.gpf.predict(model, like)
        res = self.gpf.metrics(preds[0], preds[1])
        self.gpf.store_metrics(res)
        with open(
            f"{self.gpf.input_dir}metrics_gp_cov_{self.dataset_name}_{self.version}.pickle",
            "rb",
        ) as handle:
            res = pickle.load(handle)
        keys = list(res.keys())
        self.assertTrue(keys == ['mase', 'rmse', 'CRPS', 'wall_time'])



import numpy as np
import pandas as pd


def update_missing_values(groups: dict):
    """
    Update the missing values in the 'data', 'full_data', 'data_matrix', and 'dates' arrays with np.nan

    Args:
        groups: dictionary containing the data and metadata
    """
    for key in ['train', 'predict']:
        x_values = groups[key]['x_values']
        data = groups[key]['data']
        s = groups[key]['s']

        # Change the data type of the arrays to float
        data = data.astype(float)

        # Identify the missing indices in x_values
        missing_indices = sorted(set(range(max(x_values) + 1)) - set(x_values))

        # Update the 'data' array with np.nan for the missing values
        for missing_idx in missing_indices:
            insert_idx = missing_idx * s
            data = np.insert(data, insert_idx, [np.nan] * s)

        # Update the 'n' value in the metadata
        groups[key]['n'] = len(x_values) + len(missing_indices)
        groups[key]['data'] = data

        if key == 'train':
            full_data = groups[key]['full_data']
            full_data = full_data.astype(float)

            # Update the 'full_data' array with np.nan for the missing values
            for missing_idx in missing_indices:
                insert_idx = missing_idx * s
                full_data = np.insert(full_data, insert_idx, [np.nan] * s)

            groups[key]['full_data'] = full_data

        if key == 'predict':
            data_matrix = groups[key]['data_matrix']
            data_matrix = data_matrix.astype(float)

            # Update the 'data_matrix' array with np.nan for the missing values
            for missing_idx in missing_indices:
                insert_idx = missing_idx
                if insert_idx < data_matrix.shape[0]:
                    data_matrix = np.insert(data_matrix, insert_idx, [np.nan] * s, axis=0)

            groups[key]['data_matrix'] = data_matrix

        # Update the 'dates' array to fill in missing dates
        dates = groups['dates']
        start_date = dates[0]
        end_date = dates[-1]
        all_dates = pd.date_range(start_date, end_date, freq='QS')
        missing_dates = all_dates.difference(pd.to_datetime(dates))

        for missing_date in missing_dates:
            insert_idx = all_dates.get_loc(missing_date)
            dates.insert(insert_idx, missing_date)

        groups['dates'] = dates

    return groups




import pickle
from datetime import timedelta
import os
import time
from pathlib import Path
import psutil
from typing import Dict, Union, Tuple

import torch
import numpy as np
import gpytorch
from sklearn.preprocessing import StandardScaler
from gpytorch.models import IndependentModelList
from gpytorch.likelihoods import LikelihoodList
from gpytorch.mlls import SumMarginalLogLikelihood

from htsmodels.utils.logger import Logger
from htsmodels.results.calculate_metrics import CalculateResultsBottomUp
from htsmodels.models.gp import ExactGPModel
from htsmodels.models.mean_functions import PiecewiseLinearMean, LinearMean, ZeroMean
from htsmodels import __version__


class SGP:
    def __init__(self, dataset, groups, input_dir="./", n_samples=500, log_dir="."):
        self.dataset = dataset
        self.groups = groups
        self.input_dir = input_dir
        self.timer_start = time.time()
        self.wall_time_preprocess = None
        self.wall_time_build_model = None
        self.wall_time_train = None
        self.wall_time_predict = None
        self.wall_time_total = None
        self.n_samples = n_samples

        self.groups, self.scaler = self._preprocess(groups)

        self.train_x = torch.arange(groups["train"]["n"])
        self.train_x = self.train_x.type(torch.DoubleTensor)
        self.train_x = self.train_x.unsqueeze(-1)
        self.train_y = torch.from_numpy(groups["train"]["data"])

        self.original_data_transformed = groups["predict"]["data_matrix"]

        self.n_train = groups["train"]["n"]
        self.n_predict = groups["predict"]["n"]
        self.s = groups["train"]["s"]

        self.logger_train = Logger(
            "train",
            algorithm="standard_gp",
            dataset=self.dataset,
            to_file=True,
            log_dir=log_dir,
        )
        self.logger_predict = Logger(
            "predict",
            algorithm="standard_gp",
            dataset=self.dataset,
            to_file=True,
            log_dir=log_dir,
        )
        self.logger_metrics = Logger(
            "metrics",
            algorithm="standard_gp",
            dataset=self.dataset,
            to_file=True,
            log_dir=log_dir,
        )

        self.model_version = __version__

    def _preprocess(self, groups):
        scaler = StandardScaler()
        scaler.fit(self.groups["train"]["data"])
        groups["train"]["data"] = scaler.transform(groups["train"]["data"])
        groups["predict"]["data_matrix"] = scaler.transform(
            groups["predict"]["data_matrix"]
        )

        return groups, scaler

    def _build_cov_matrices(self):
        covs = []
        for i in range(self.groups["train"]["s"]):
            # RBF kernel
            rbf_kernel = gpytorch.kernels.RBFKernel()
            rbf_kernel.lengthscale = torch.tensor([1.0])
            scale_rbf_kernel = gpytorch.kernels.ScaleKernel(rbf_kernel)
            scale_rbf_kernel.outputscale = torch.tensor([0.5])

            # Periodic Kernel
            periodic_kernel = gpytorch.kernels.PeriodicKernel()
            periodic_kernel.period_length = torch.tensor([self.groups["seasonality"]])
            periodic_kernel.lengthscale = torch.tensor([0.5])
            scale_periodic_kernel = gpytorch.kernels.ScaleKernel(periodic_kernel)
            scale_periodic_kernel.outputscale = torch.tensor([1.5])

            # Cov Matrix
            cov = scale_rbf_kernel + scale_periodic_kernel
            covs.append(cov)

        return covs

    def _build_model(self, mean_function):
        covs = self._build_cov_matrices()
        n_changepoints = 4
        changepoints = np.linspace(0, self.groups["train"]["n"], n_changepoints + 2)[
            1:-1
        ]

        model_list = []
        likelihood_list = []
        for i in range(self.groups["train"]["s"]):
            likelihood_list.append(gpytorch.likelihoods.GaussianLikelihood())
            if mean_function == "piecewise_linear":
                model_list.append(
                    ExactGPModel(
                        self.train_x,
                        self.train_y[:, i],
                        likelihood_list[i],
                        covs[i],
                        PiecewiseLinearMean(changepoints),
                    )
                )
            elif mean_function == "zero":
                model_list.append(
                    ExactGPModel(
                        self.train_x,
                        self.train_y[:, i],
                        likelihood_list[i],
                        covs[i],
                        gpytorch.means.ZeroMean(),
                    )
                )
            elif mean_function == "linear":
                model_list.append(
                    ExactGPModel(
                        self.train_x,
                        self.train_y[:, i],
                        likelihood_list[i],
                        covs[i],
                        LinearMean(1),
                    )
                )

        return likelihood_list, model_list

    def train(
        self,
        n_iterations=150,
        lr=1e-3,
        mean_function="piecewise_linear",
        track_mem=True,
    ):
        likelihood_list, model_list = self._build_model(mean_function)

        model = gpytorch.models.IndependentModelList(*model_list)
        likelihood = gpytorch.likelihoods.LikelihoodList(*likelihood_list)

        mll = SumMarginalLogLikelihood(likelihood, model)

        model.train()
        likelihood.train()

        optimizer = torch.optim.Adam(
            model.parameters(), lr=lr
        )  # Includes GaussianLikelihood parameters

        for i in range(n_iterations):
            optimizer.zero_grad()
            output = model(*model.train_inputs)
            loss = -mll(output, model.train_targets)
            loss.backward()
            print("Iter %d/%d - Loss: %.3f" % (i + 1, n_iterations, loss.item()))
            optimizer.step()

            if i % 30 == 0 and track_mem:
                # Track RAM usage
                process = psutil.Process(os.getpid())
                mem = process.memory_info().rss / (1024**3)
                self.logger_train.info(f"train used {mem:.3f} GB of RAM")

        self.wall_time_train = time.time() - self.timer_start
        td = timedelta(seconds=int(time.time() - self.timer_start))
        self.logger_train.info(f"Num epochs {i}")
        self.logger_train.info(f"wall time train {str(td)}")

        return model, likelihood

    def predict(
        self,
        model: IndependentModelList,
        likelihood: LikelihoodList,
        track_mem: bool = True,
    ) -> Tuple[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]:
        """
        Make predictions with the model.

        Parameters:
            model: The GP model.
            likelihood: The likelihood function.
            track_mem: Track and log RAM usage

        Returns:
            numpy.ndarray: Array of shape (n_samples, n_prediction_points, n_groups)
                containing the prediction samples.
        """
        timer_start = time.time()

        model.eval()
        likelihood.eval()

        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            test_x = torch.arange(self.groups["predict"]["n"]).type(torch.DoubleTensor)
            predictions = likelihood(
                *model(*[test_x for i in range(self.groups["predict"]["s"])])
            )
            if track_mem:
                # Track RAM usage
                process = psutil.Process(os.getpid())
                mem = process.memory_info().rss / (1024**3)
                self.logger_predict.info(f"predict used {mem:.3f} GB of RAM")

        pred_mean_scaled = np.zeros((self.n_predict, self.s))
        pred_std_scaled = np.zeros((self.n_predict, self.s))
        for ts in range(self.s):
            pred_mean_scaled[:, ts] = predictions[ts].mean.detach().numpy()
            pred_std_scaled[:, ts] = np.sqrt(predictions[ts].variance.detach().numpy())

        # transform back the data
        pred_mean = pred_mean_scaled * self.scaler.scale_ + self.scaler.mean_
        pred_std = pred_std_scaled * self.scaler.scale_

        self.wall_time_predict = time.time() - timer_start
        return (pred_mean, pred_std), (pred_mean_scaled, pred_std_scaled)

    @staticmethod
    def _validate_param(param, valid_values):
        if param not in valid_values:
            raise ValueError(f"{param} is not a valid value")

    def store_results(
        self,
        res: np.ndarray,
        res_type: str,
        res_measure: str,
        track_mem: bool = True,
    ):
        """
        Store results

        Parameters:
            res: np array with the results with shape (n,s) - note that n depends of the res_type
            res_type: defines the type of results, could be 'fit_pred' to receive fitted values plus
                predictions or 'pred' to only store predictions
            res_measure: defines the measure to store, could be 'mean' or 'std'

        Returns:
            numpy.ndarray: Array of shape (n_samples, n_prediction_points, n_groups)
                containing the prediction samples.
        """
        """
        Store results, res_type should be used to define the type of results,
        could be 'fit_pred' to receive fitted values plus predictions or 'pred'
        to only store predictions
        """
        self._validate_param(res_type, ['fitpred', 'pred'])
        self._validate_param(res_measure, ['mean', 'std'])
        with open(
            f"{self.input_dir}results_{res_type}_{res_measure}_gp_cov_{self.dataset}_{self.model_version}.pickle",
            "wb",
        ) as handle:
            if track_mem:
                process = psutil.Process(os.getpid())
                mem = process.memory_info().rss / (1024**3)
                self.logger_metrics.info(f"Storing results used {mem:.3f} GB of RAM")
            pickle.dump(res, handle, pickle.HIGHEST_PROTOCOL)

    def store_metrics(
        self,
        res: Dict[str, Dict[str, Union[float, np.ndarray]]],
        track_mem: bool = True,
    ):
        with open(
            f"{self.input_dir}metrics_gp_cov_{self.dataset}_{self.model_version}.pickle",
            "wb",
        ) as handle:
            if track_mem:
                process = psutil.Process(os.getpid())
                mem = process.memory_info().rss / (1024**3)
                self.logger_metrics.info(
                    f"Storing error metrics used {mem:.3f} GB of RAM"
                )
            pickle.dump(res, handle, pickle.HIGHEST_PROTOCOL)

    def metrics(
        self,
        pred_mean: np.ndarray,
        pred_std: np.ndarray,
        track_mem: bool = True,
    ) -> Dict[str, Dict[str, Union[float, np.ndarray]]]:
        """
        Calculate evaluation metrics for the predictions.

        Parameters:
            pred_mean: Array of shape (n_prediction_points, n_series)
                containing the prediction samples.
            pred_std: Array of shape (n_prediction_points, n_series)
                containing the prediction samples.
            track_mem: Track and log RAM usage

        Returns:
            dict: Dictionary with the evaluation metrics. The keys are the metric names,
                and the values are dictionaries with the results for each group.
        """
        calc_results = CalculateResultsBottomUp(
            predictions_mean=pred_mean,
            predictions_std=pred_std,
            groups=self.groups,
            dataset=self.dataset,
        )
        res = calc_results.calculate_metrics()
        if track_mem:
            process = psutil.Process(os.getpid())
            mem = process.memory_info().rss / (1024**3)
            self.logger_metrics.info(
                f"calculating error metrics used {mem:.3f} GB of RAM"
            )
        self.wall_time_total = time.time() - self.timer_start

        res["wall_time"] = {}
        res["wall_time"]["wall_time_train"] = self.wall_time_train
        res["wall_time"]["wall_time_predict"] = self.wall_time_predict
        res["wall_time"]["wall_time_total"] = self.wall_time_total

        return res
